# coding=utf_8
'''
    PGExt - pygame extension
    Created by Josef Vanžura <gindar@zamraky.cz>
'''

__version__ = "1.13"

import filters
import color

from filters import *
from color import *

__version_filters__ = filters.version
__version_color__ = color.version
